module ani_parms

  use sep
  use ani_type

  implicit none

  ! Forward or adjoint, dottest or not
  logical,        public, save				:: adj,dot

  ! Define the parameter grid given by input data
  type (dat_grid), public, save	        		:: rec, sou
  type (img_grid), public, save	        		:: img
  type (modl) , public, save				:: mdl, dmdl
  type (mref) , public, save                            :: ref
  type (axis) , public, save     			:: h, w, xd, xm, z, sh, kz, ksh, th, par, rf

  ! Define the useful parameters in the program
  type (axis),    public, save  			:: kxm
  integer	 ,public, save				:: iwb,tapertr,padtraces
  integer	 ,public, save				:: isp,node,nsecdim
  integer	 ,public, save				:: nx, nz, nt, nsh, npar, nref_max
  real   	 ,public, save				:: dx, ox, dz, oz, dt, ot, dsh, osh
  real           ,public, save				:: pi, osecdim, dsecdim
  real, allocatable, public, save                       :: scales(:)

contains

  subroutine aniparam()

    integer						:: iw, i

    ! ----------------------------------------------------------------------------------------------------------------------
    ! Read in parameters
    call from_param("dot",dot,.false.)
    call from_param("adj",adj,.true.)

    call from_param("node",node,9999)
    call from_param("traces_in_taper",tapertr,0)
    call from_param("padtraces",padtraces,tapertr)

    ! Read the receiver parameter. xd is the shot position
    call from_aux("rec","n1",rec%h%n);	call from_aux("rec","d1",rec%h%d);call from_aux("rec","o1",rec%h%o)
    call from_aux("rec","n2",rec%w%n);	call from_aux("rec","d2",rec%w%d);call from_aux("rec","o2",rec%w%o)
    call from_aux("rec","n3",rec%xd%n);call from_aux("rec","d3",rec%xd%d);call from_aux("rec","o3",rec%xd%o)  

    ! Read the source parameter. 
    call from_aux("sou","n1",sou%w%n);	call from_aux("sou","d1",sou%w%d);	call from_aux("sou","o1",sou%w%o)

    ! Read the model parameters.
    call from_aux("mdl","n1",mdl%z%n );	call from_aux("mdl","d1",mdl%z%d );	call from_aux("mdl","o1",mdl%z%o )
    call from_aux("mdl","n2",mdl%xm%n);	call from_aux("mdl","d2",mdl%xm%d);	call from_aux("mdl","o2",mdl%xm%o)
    call from_aux("mdl","n3",mdl%par%n);	

    ! Read the reference model parameters.
    call from_aux("ref","n1",ref%z%n );	call from_aux("ref","d1",ref%z%d );	call from_aux("ref","o1",ref%z%o )
    call from_aux("ref","n2",ref%rf%n);	
    call from_aux("ref","n3",ref%par%n); 	

    ! Either read in the image or slowness pertubation. 
    ! For image, the third axis is the shot position.
    if (adj) then
       call from_aux("deltaI","n1",img%z%n ); call from_aux("deltaI","d1",img%z%d) ; call from_aux("deltaI","o1",img%z%o )  
       call from_aux("deltaI","n2",img%xm%n); call from_aux("deltaI","d2",img%xm%d); call from_aux("deltaI","o2",img%xm%o)
       call from_aux("deltaI","n3",img%sh%n); call from_aux("deltaI","d3",img%sh%d); call from_aux("deltaI","o3",img%sh%o)
    else
       call from_aux("deltaM","n1",dmdl%z%n );	call from_aux("deltaM","d1",dmdl%z%d );	call from_aux("deltaM","o1",dmdl%z%o )
       call from_aux("deltaM","n2",dmdl%xm%n);	call from_aux("deltaM","d2",dmdl%xm%d);	call from_aux("deltaM","o2",dmdl%xm%o)
       call from_aux("deltaM","n3",dmdl%par%n);	
    end if



    ! ----------------------------------------------------------------------------------------------------------------------
    ! Check the coherence of model and reference model
    if (mdl%z%n .ne. ref%z%n) then
       call seperr("# of Model depth .ne. # of reference model depth")
    end if
    if (mdl%par%n .ne. ref%par%n) then
       call seperr("# of Model par. .ne. # of reference model par.")
    end if

    ! Check the coherence of the source and receiver
    if (sou%w%n .ne. rec%w%n) then
       call seperr("# source freq. .ne. # receiver freq.")
    end if
    if (sou%w%d .ne. rec%w%d) then
       call seperr("source freq.interval .ne. receiver freq.interval")
    end if
    if (sou%w%o .ne. rec%w%o) then
       call seperr("source freq.origin .ne. receiver freq.origin")
    end if

    ! Check the coherence of deltaM and model
    if (not(adj)) then
       if (dmdl%z%n  .ne. mdl%z%n) then
          call seperr("deltaM # depth samples .ne. model  ")
       end if
       if (dmdl%xm%n .ne. mdl%xm%n) then
          call seperr("deltaM # traces .ne. model  ")
       end if
       if (dmdl%par%n .ne. mdl%par%n) then
          call seperr("deltaM # par. .ne. model par. ")
       end if
    end if




    ! ----------------------------------------------------------------------------------------------------------------------
    ! Compute the useful parameters
    pi=4*atan(1.)
    nref_max = ref%rf%n; npar = ref%par%n;
    nsh=rec%xd%n; dsh=rec%xd%d; osh=rec%xd%o
    write(0,*) "nsh=", nsh, "dsh=", dsh, "osh=", osh


    ! ----------------------------------------------------------------------------------------------------------------------
    ! The final 2-D grid that we are working on
    nz = mdl%z%n; dz = mdl%z%d; oz = mdl%z%o
    nx = rec%h%n+2*padtraces+2*tapertr  
    dx = rec%h%d
    ox = rec%h%o-(padtraces+tapertr)*dx 
    img%xm%n=nx; img%xm%o=ox;img%xm%d=dx
    img%z%d=dz
    if (nx .ne. mdl%xm%n ) then
       write(0,*) "nx=", nx, "mdl%xm%n=", mdl%xm%n
       call seperr("# of working nx .ne. # of model nx")
    end if
    if (ox .ne. mdl%xm%o ) then
       write(0,*) ox, mdl%xm%o
       call seperr("Min.x-image space .ne. Min.x-model")
    end if
    if (dx .ne. mdl%xm%d ) call seperr("Sampling of working grid .ne. Sampling of model gird")    


    allocate(scales(npar))
    call from_par("scales",scales)

    !--------------------------------------------------
    ! Subsurface offset vector for the image
    !
    call from_param("nh",nsecdim,1)
    call from_param("dh",dsecdim,img%xm%d)
    call from_param("oh",osecdim,-int(nsecdim/2)*dsecdim)


    !--------------------------------------------------
    ! Spatial Frequency vector
    !
    kxm%n = rec%h%n+2*(tapertr+padtraces)
    kxm%d = 1./(kxm%n*rec%h%d)
    kxm%o = 1./(2*rec%h%d)-kxm%d
    allocate(kxm%all(kxm%n)); kxm%all=0.
    do i=1,kxm%n/2+1
       kxm%all(i) = (i-1) * kxm%d
    end do
    do i=kxm%n/2+2, kxm%n
       kxm%all(i) = (i-kxm%n-1) * kxm%d
    end do
    kxm%all = kxm%all * pi*2

    !--------------------------------------------------
    ! Frequency vector
    !
    rec%w%o   = 2*pi * rec%w%o 
    rec%w%d   = 2*pi * rec%w%d  
    allocate(rec%w%all(rec%w%n))
    do iwb = 1,rec%w%n
       rec%w%all(iwb) = rec%w%o + (iwb - 1) * rec%w%d
    end do

    !--------------------------------------------------
    ! Time axis
    ! FFt from frequency
    nt = rec%w%n * 2
    dt = 1./(rec%w%n*nt) * pi*2
    ot = 0.

!!!!!!!!!!!!!!!!!
    !
    ! works for symmetric offset

    if (adj) then
       ! dimension for model updates
       call to_history("n1",nz) ; call to_history("d1",dz) ; call to_history("o1",oz)
       call to_history("n2",nx) ; call to_history("d2",dx) ; call to_history("o2",ox)
       call to_history("n3",npar) ; call to_history("d3",1 ) ; call to_history("o3",0 )
       call to_history("label1", "z(m)"); call to_history("label2", "x(m)"); call to_history("label3", "par"); 
       call to_history("esize",4)
    else
       ! dimension for image purtubations
       call to_history("n1",nz) ; call to_history("d1",dz) ; call to_history("o1",oz)
       call to_history("n2",nx) ; call to_history("d2",dx) ; call to_history("o2",ox)
       call to_history("n3",nsecdim) ; call to_history("d3",dsecdim) ; call to_history("o3",osecdim)
       call to_history("n4",1); call to_history("d4",1); call to_history("o4",0)
       call to_history("n5",1); call to_history("d5",1); call to_history("o5",0)  
       call to_history("esize",4)
    end if
  end subroutine aniparam
end module ani_parms
