module ani_wemva

  use sep
  use ani_type
  use ani_fftw
  use ani_parms
  use ani_image
  use eta_pspi
  use eta_phsscr

  implicit none
 
  integer, private                          :: idip, nodes

contains

  !======================================================================
  subroutine aniwei_init()
    integer, external			:: omp_get_thread_num,omp_get_num_threads
    if (node.eq.9999) then
       !$OMP PARALLEL
       nodes = omp_get_num_threads()
       !$OMP END PARALLEL
    else
       nodes=node
    end if
    write(3,*) "****************"
    write(3,*) "* NUM.THREADS = ",nodes
    write(3,*) "****************"
    
  end subroutine aniwei_init

  !======================================================================
  subroutine aniwei(taper,recv,source,model,refr,image)

    complex						:: recv(:,:),source(:,:)
    integer						:: iwa,iz,inode,i,ixw,ish
    real						:: model(:,:,:), refr(:,:,:)
    real						:: image(:,:,:), taper(:)
    real, allocatable					:: GG(:,:,:,:)
    complex, allocatable				:: ws(:,:),wr(:,:)
    integer, external   			        :: omp_get_thread_num,omp_get_num_threads
    
    allocate(ws(nx,nodes), wr(nx,nodes))
    wr=0
    ws=0
    allocate(GG(nz,nx,nsecdim,nodes))
    GG=0.
    !$OMP parallel do private(inode,iwa,iz) ordered schedule(dynamic,1) num_threads(nodes)
    do iwa = 2,rec%w%n
       inode=omp_get_thread_num()+1;
       ws(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode) = source(:,iwa)
       wr(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode) =   recv(:,iwa)
       call aniimage(.false., iwa, ws(:,inode), wr(:,inode), GG(1,:,:,inode))          
       do iz = 2, nz
          call etapspi(iwa, inode, -1, ws(:,inode), model(iz-1,:,:), refr(iz-1,:,:), "sou")
          call etapspi(iwa, inode,  1, wr(:,inode), model(iz-1,:,:), refr(iz-1,:,:), "rec")
          ws(:,inode)=ws(:,inode)*taper
          wr(:,inode)=wr(:,inode)*taper
          call aniimage(.false., iwa, ws(:,inode), wr(:,inode), GG(iz,:,:,inode))
       end do
       if (mod(iwa,50)==1) write(3,6) " FREQUENCY # ",iwa," OUT OF ",rec%w%n," / SYNTH.FCT. # ",rec%xd%n, &
            " FROM NODE: ",inode
    end do
    !$OMP END PARALLEL DO

    do i=1,nodes
       !$OMP parallel do private(ixw,iz,ish)
       do iz=1,nz
          do ixw=1,kxm%n
             do ish=1,nsecdim
                image(iz,ixw,ish)=image(iz,ixw,ish)+GG(iz,ixw,ish,i)
             end do
          end do
       end do
       !$OMP END PARALLEL DO
    end do
6   format(A13,i3,A8,i3,A14,i3,A12,i3)
  end subroutine aniwei


  !======================================================================
  subroutine aniwemva_init(idip_in)

    integer      :: idip_in, idip
    integer, external			:: omp_get_thread_num,omp_get_num_threads
    if (node.eq.9999) then
       !$OMP PARALLEL
       nodes = omp_get_num_threads()
       !$OMP END PARALLEL
    else
       nodes=node
    end if

    write(3,*) "****************"
    write(3,*) "* NUM.THREADS = ",nodes
    write(3,*) "****************"
    
    idip = idip_in
  end subroutine aniwemva_init
  !======================================================================
  


!======================================================================
  subroutine aniwemva(adj,taper,recv,source,model,refr,deltaM,deltaI)

    logical						:: adj
    complex						:: recv(:,:),source(:,:)
    integer						:: iwa,iz,inode,i,ixw,ish
    real						:: model(:,:,:), refr(:,:,:)
    real						:: deltaM(:,:,:)
    real						:: deltaI(:,:,:), taper(:)
    real, allocatable					:: SSS(:,:,:), EEE(:,:,:)
    real, allocatable					:: GG(:,:,:,:)
    complex, allocatable				:: ws(:,:),wr(:,:)
    complex, allocatable				:: ddws(:,:),ddwr(:,:)
    complex, allocatable				:: dws(:,:,:),dwr(:,:,:)
    complex, allocatable				:: bws(:,:,:),bwr(:,:,:)
    integer, external   			        :: omp_get_thread_num,omp_get_num_threads
    
    allocate(bws(nx,nz,nodes), bwr(nx,nz,nodes))
    allocate(ws(nx,nodes), wr(nx,nodes))
    allocate(ddws(nx,nodes), ddwr(nx,nodes))
    wr=0
    ws=0
    ddwr=0
    ddws=0
    bwr=0
    bws=0

    ! ----------------------------------------
    ! Adjoint tomographic operator
    if (adj) then
       allocate(SSS(nx,nz,nodes),EEE(nx,nz,nodes))
       deltaM=0.
       SSS=0.                                 ! Temporary location for slowness perturbation
       EEE=0.                                 ! Temporary location for \eta perturbation 
       allocate(dwr(1,kxm%n,nodes));     dwr=0.
       allocate(dws(1,kxm%n,nodes));     dws=0.

       !$OMP parallel do private(inode,iwa,iz) ordered schedule(dynamic,1) num_threads(nodes)       
       do iwa = 2, rec%w%n
          ! Initializing 
          inode=omp_get_thread_num()+1; 
          ws(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode) = source(:,iwa)     ! Source and recv are not padded
          wr(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode) =   recv(:,iwa)
          dws(1,:,inode) = 0.
          dwr(1,:,inode) = 0.
          bws(:,:,inode) = 0.
          bwr(:,:,inode) = 0.
          bws(:,1,inode) = ws(:,inode)
          bwr(:,1,inode) = wr(:,inode)

          ! Background wavefields, bws & bwr
          do iz = 2, nz
             call etapspi(iwa, inode, -1, ws(:,inode), model(iz-1,:,:), refr(iz-1,:,:),"sou")
             call etapspi(iwa, inode,  1, wr(:,inode), model(iz-1,:,:), refr(iz-1,:,:),"rec")
             ws(:,inode)=ws(:,inode)*taper
             wr(:,inode)=wr(:,inode)*taper
             bws(:,iz,inode)=ws(:,inode)
             bwr(:,iz,inode)=wr(:,inode)
          end do
          
          do iz=nz, 1, -1
             ! Adjoint imaging condition to get the wavefield perturbation
             call aniimage(.true., iwa, dws(1,:,inode), bwr(:,iz,inode), deltaI(iz,:,:),  .true.)
             call aniimage(.true., iwa, bws(:,iz,inode), dwr(1,:,inode), deltaI(iz,:,:), .false.)
!	     write(0,*) iz, "dws,", sum(abs(dws(1,:,inode))), "dwr,", sum(abs(dwr(1,:,inode))), "deltaI", sum(deltaI(iz,:,:))

             ! Upward propagate wavefield perturbations
             call etapspi(iwa, inode,  1, dws(1,:,inode), model(iz,:,:), refr(iz,:,:), "sou")
             dws(1,:,inode)=dws(1,:,inode)*taper
             call etapspi(iwa, inode, -1, dwr(1,:,inode), model(iz,:,:), refr(iz,:,:), "rec")
             dwr(1,:,inode)=dwr(1,:,inode)*taper
             
             ! Phase-screen operator & adjoint scattering operator to get the model perturbation
             ddws(:,inode)=0
             call aniphsscr_slow(iwa,-1,ddws(:,inode),bws(:,iz,inode),model(iz,:,1),model(iz,:,2))
             SSS(:,iz,inode)=SSS(:,iz,inode)-conjg(ddws(:,inode))*(dws(1,:,inode))/rec%w%n    
!            write(0,*) "SSS, ", sum(abs(ddws(:,inode))), sum(abs(conjg(ddws(:,inode))*(dws(1,:,inode))))
             
             ddws(:,inode)=0
             call aniphsscr_eta (iwa,-1,ddws(:,inode),bws(:,iz,inode),model(iz,:,1))
             EEE(:,iz,inode)=EEE(:,iz,inode)-conjg(ddws(:,inode))*(dws(1,:,inode))/rec%w%n    
!             write(0,*) iz, sum(abs(ddws(:,inode))), sum(abs(conjg(ddws(:,inode))*(dws(1,:,inode))))
             
             ddwr(:,inode)=0
             call aniphsscr_slow(iwa, 1,ddwr(:,inode),bwr(:,iz,inode),model(iz,:,1),model(iz,:,2))
             SSS(:,iz,inode)=SSS(:,iz,inode)-conjg(ddwr(:,inode))*(dwr(1,:,inode))/rec%w%n
             
             ddwr(:,inode)=0
             call aniphsscr_eta (iwa, 1,ddwr(:,inode),bwr(:,iz,inode),model(iz,:,1))
             EEE(:,iz,inode)=EEE(:,iz,inode)-conjg(ddwr(:,inode))*(dwr(1,:,inode))/rec%w%n  
          end do

          if (mod(iwa,50)==1) write(3,6) " FREQUENCY # ",iwa," OUT OF ",rec%w%n," / SYNTH.FCT. # ",rec%xd%n, &
               " FROM NODE: ",inode
       end do
       !$OMP END PARALLEL DO
   
       do i=1,nodes
          !$OMP parallel do private(ixw,iz)
          do iz=1,nz
             do ixw=1,kxm%n
                deltaM(iz,ixw,1)=deltaM(iz,ixw,1)+SSS(ixw,iz,i)
                deltaM(iz,ixw,2)=deltaM(iz,ixw,2)+EEE(ixw,iz,i)
             end do
          end do
          !$OMP END PARALLEL DO
       end do
       deallocate(dwr,dws,EEE,SSS)
    ! ----------------------------------------
    ! Forward tomographic operator
    else 
       allocate(GG(nz,nx,nsecdim,nodes))
       deltaI=0.
       GG=0.
       allocate(dwr(1,kxm%n,nodes));               dwr=0.
       allocate(dws(1,kxm%n,nodes));               dws=0.
       !$OMP parallel do private(inode,iwa,iz) ordered schedule(dynamic,1) num_threads(nodes)

       do iwa = 2,rec%w%n
          inode=omp_get_thread_num()+1;
          ws(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode) = source(:,iwa)
          wr(tapertr+padtraces+1:rec%h%n+padtraces+tapertr,inode) =   recv(:,iwa)
          dws(1,:,inode)=0.
          dwr(1,:,inode)=0.
          bws(:,:,inode)=0.
          bwr(:,:,inode)=0.
          bws(:,1,inode)=ws(:,inode)
          bwr(:,1,inode)=wr(:,inode)

          do iz = 2, nz
             call etapspi(iwa, inode, -1, ws(:,inode), model(iz-1,:,:), refr(iz-1,:,:), "sou")
             call etapspi(iwa, inode,  1, wr(:,inode), model(iz-1,:,:), refr(iz-1,:,:), "rec")
             ws(:,inode)=ws(:,inode)*taper
             wr(:,inode)=wr(:,inode)*taper
             bws(:,iz,inode)=ws(:,inode)
             bwr(:,iz,inode)=wr(:,inode)
          end do
          
          do iz=1,nz
             ! Phase-screen
             ddws(:,inode) = 0
             call aniphsscr_slow(iwa, 1,ddws(:,inode),bws(:,iz,inode),model(iz,:,1),model(iz,:,2))
             dws(1,:,inode)=dws(1,:,inode)+(ddws(:,inode))*deltaM(iz,:,1)/rec%w%n   
             
             ddws(:,inode)=0
             call aniphsscr_eta (iwa, 1,ddws(:,inode),bws(:,iz,inode),model(iz,:,1))
             dws(1,:,inode)=dws(1,:,inode)+(ddws(:,inode))*deltaM(iz,:,2)/rec%w%n    

             ! Scattering
             call etapspi(iwa, inode,-1, dws(1,:,inode), model(iz,:,:), refr(iz,:,:),"sou")
             dws(1,:,inode)=dws(1,:,inode)*taper
!	     write(0,*) iz, "dws,", sum(abs(dws(1,:,inode))), "dwr,", sum(abs(dwr(1,:,inode)))
             
             ! Imaging condition
             call aniimage(.false., iwa, dws(1,:,inode), bwr(:,iz,inode), GG(iz,:,:,inode))

             ! Phase-screen
             ddwr(:,inode)=0
             call aniphsscr_slow(iwa,-1,ddwr(:,inode),bwr(:,iz,inode),model(iz,:,1),model(iz,:,2))
             dwr(1,:,inode)=dwr(1,:,inode)+(ddwr(:,inode))*deltaM(iz,:,1)/rec%w%n    

             ddwr(:,inode)=0
             call aniphsscr_eta (iwa,-1,ddwr(:,inode),bwr(:,iz,inode),model(iz,:,1))
             dwr(1,:,inode)=dwr(1,:,inode)+(ddwr(:,inode))*deltaM(iz,:,2)/rec%w%n 

             ! Scattering
             call etapspi(iwa, inode, 1, dwr(1,:,inode), model(iz,:,:), refr(iz,:,:),"rec")
             dwr(1,:,inode)=dwr(1,:,inode)*taper
             
             ! Imaging condition
             call aniimage(.false., iwa, bws(:,iz,inode), dwr(1,:,inode), GG(iz,:,:,inode))
          end do

          if (mod(iwa,50)==1) write(3,6) " FREQUENCY # ",iwa," OUT OF ",rec%w%n," / SYNTH.FCT. # ",rec%xd%n, &
               " FROM NODE: ",inode
       end do
       !$OMP END PARALLEL DO

       do i=1,nodes
          !$OMP parallel do private(ixw,iz,ish)
          do iz=1,nz
            do ixw=1,kxm%n
             do ish=1,nsecdim
                   deltaI(iz,ixw,ish)=deltaI(iz,ixw,ish)+GG(iz,ixw,ish,i)
                end do
             end do
          end do
          !$OMP END PARALLEL DO
       end do
       deallocate(dwr,dws,GG)

    end if
 
    deallocate(bws,bwr)
6   format(A13,i3,A8,i3,A14,i3,A12,i3)
  end subroutine aniwemva
  !==================================================
end module ani_wemva

