module eta_phsscr

use ani_parms

implicit none

contains
!
! Anisotropic phase screen
!
subroutine aniphsscr_slow(iwa,sign,ddws,wv,slow,eta)

complex                 :: ddws(:), wv(:)
real	        	:: slow(:), eta(:)
real                    :: dx2, w2, s2, w4, s4
integer                 :: sign,ix,iwa
complex                 :: iwdz, lap2(kxm%n), lap4(kxm%n)
real                    :: a(kxm%n), b(kxm%n)

a = 0.5
b = 2*eta + 0.25
lap2=0; lap4=0;

   if (rec%w%all(iwa).ne.0.) then
      iwdz = cmplx(0., 1.)*rec%w%all(iwa)*img%z%d*sign 
      dx2  = img%xm%d**2
      w2   = rec%w%all(iwa)**2
      w4   = w2**2
      call laplac2(kxm%n, img%xm%d, wv  , lap2)
      call laplac2(kxm%n, img%xm%d, lap2, lap4)
      do ix=1, kxm%n
         s2 = slow(ix)
         s2=s2*s2
         s4=s2*s2
         ddws(ix) = iwdz*(wv(ix)+a(ix)/w2/s2*lap2(ix)+3*a(ix)*b(ix)/w4/s4*lap4(ix)) 
      end do
   else
      ddws=0.
   end if

 end subroutine aniphsscr_slow
 

subroutine aniphsscr_eta(iwa,sign,ddws,wv,slow)

complex                 :: ddws(:), wv(:)
real	        	:: slow(:)
real                    :: dx2, w2, s2, w4, s4
integer                 :: sign,ix,iwa
complex                 :: iwdz, lap2(kxm%n), lap4(kxm%n)

   lap2=0; lap4=0;

   if (rec%w%all(iwa).ne.0.) then
      iwdz = cmplx(0., 1.)*rec%w%all(iwa)*img%z%d*sign 
      dx2  = img%xm%d**2
      w2   = rec%w%all(iwa)**2
      w4   = w2**2
      call laplac2(kxm%n, img%xm%d, wv  , lap2)
      call laplac2(kxm%n, img%xm%d, lap2, lap4)

      do ix=1, kxm%n
         s2 = slow(ix); s2=s2*s2; s4=s2*s2
         ddws(ix) = -slow(ix)*iwdz/w4/s4*lap4(ix) 
      end do
   else
      ddws=0.
   end if

 end subroutine aniphsscr_eta


!!--------------------------------------------------------------------------
!! 2nd Order Laplacian

  subroutine laplac2(nx,dx, uo,ud)

    integer    :: ix,nx
    real       :: dx
    complex, dimension(:)  :: uo,ud

       ud(1)  = (uo(2) - uo(1))/dx
       ud(nx) = (uo(nx) - uo(nx-1))/dx
       do ix=1+1,nx-1
          ud(ix) = &
               (-2.*uo(ix) + uo(ix-1) + uo(ix+1))/dx**2 
       end do
  end subroutine 

end module eta_phsscr
